#include<fstream>
#include<iostream>
#include<sstream>
#include<vector>

using namespace std;

#include "indexer.h"

indexer::indexer() {
	for(int i = 0; i < INDEXSIZE; i++)
		geneIndex[i] = new vector<int>();

	//FIXME: this must be changed when funciton setting up mapper is created; should probably actually just use an array
	LTscore.resize(55);
}

int indexer::hash(const string key) {
	int rv = 0;
	int base = 1;

	for(int i = KEYSIZE - 1; i >= 0; i--) {
		switch(key[i]) {
			case 'A':
				break;
			case 'C':
				rv += base;
				break;
			case 'G':
				rv += 2*base;
				break;
			case 'T':
				rv += 3*base;
				break;
			default:
				cerr << "Input to hash function is invalid." << endl;
				break;
		}

		base *= 4;
	}

	return rv;
}

int indexer::index(const string filename, const string testCases) {
	
	ifstream ref;	//stream to use with reference genome
	char buf[BUFSIZE + KEYSIZE];		//input buffer
	int iNum;
	unsigned int count(0);
	string key, fileName;
	vector<int>::iterator iter, end;

	//open up file
	ref.open(filename.c_str());

	if(!ref.good()) {
		cerr << "Error: Cannot open " << filename << "for indexing!" << endl;
		return 1;
	}
	
	int i = KEYSIZE - 1;	//want to do first indexing with an offset

	//read in 1024 bytes of file at a time
	while(ref.good()) {	
		for(int k = 0; k < 10000; ++k) {
			ref.get(buf + KEYSIZE - 1, BUFSIZE + 1);

			if(!ref.bad()) {
					
				//for each KEYSIZE bytes, hash them in geneIndex and store the index of the reference genome as the value
				for(; buf[i + KEYSIZE - 1] != '\0'; i++) {
					key = string(buf + i, KEYSIZE);		//create string

					iNum = hash(key);						//new indexing with my hash function
					geneIndex[iNum]->push_back(count);
						
					count++;		//increment index
				}						

			}

			for(int j = 0; j < KEYSIZE - 1; j++) {	//move extra nucleotides from end of buffer to beginning of buffer
				buf[j] = buf[j + i];
			}
			
			i = 0;		//begin next indexing from beginning of buffer	
		} //index for these 10M nucleotides is complete

		//FIXME: debugging statement
		cerr << "At index " << count << endl;

		//skip seeker back 30
		ref.seekg(-30, ios_base::cur);
		count -= 30;

		//TODO: make it so that map doesn't have to read reads from file every time (they should be stored using another function)

		map(testCases);

		//clear index for next files
		for(int m = 0; m < INDEXSIZE; m++) {
			geneIndex[m]->clear();
		}

	}

		//close file
		ref.close();

		return 0;
}

int indexer::map(const std::string read) {
	
	ifstream inf;		//open up stream for read file and write file
	char buf[256];

	//FIXME: iLength should be taken from LTscore
	int iLength(30), iNum, readNum(0);
	string key;
	//hashmap::iterator giIter;
	//pair<hashmap::iterator, hashmap::iterator> bounds;
	vector<int>::iterator giIter, bound;
	tallier::iterator mapKey;

	inf.open(read.c_str());

	if(!inf.good()) {
		cerr << "Error: Cannot open " << read << "for input!" << endl;
		return 1;
	}

beginMap:
	while(inf.good()) {		//while not EOF for read
		indexScore.clear();

		inf.getline(buf, 256); //read in a line to buffer

		if(inf.bad()) {
			cerr << "Error reading from " << read << "for input!" << endl;
			return 1;
		}

		for(int i = 0; i <= iLength - KEYSIZE; i += 7) {		//check the 4 key positions
			key = string(buf + i, KEYSIZE);		//hash 8 bytes of our string
			
			iNum = hash(key);
			//bounds = geneIndex.equal_range(key);		//old indexer
			
			for(giIter = geneIndex[iNum]->begin(), bound = geneIndex[iNum]->end(); giIter!=bound; ++giIter) { //iterate through indices
				mapKey = indexScore.find(*giIter - i);				//find out if index (minus i) is already in map
					
				if(mapKey == indexScore.end()) {		 //if they aren't in map, insert them with value = 1
					indexScore.insert(make_pair(*giIter - i, 1));
				}
				else {				//if they are in map, add 1 to their value
					if(mapKey->second == 2) {			//if we have 3 hits, declare this index the winner
						LTscore[readNum].index = mapKey->first;
						LTscore[readNum].quality = 100;
						readNum++;
						goto beginMap;
					}
					mapKey->second++;
				}
			}
		}

		for(int i = 1; i <= iLength - KEYSIZE; ++i) {		//until we reach the size of the substring - KEYSIZE + 1
			if(!(i%7))
				++i;

			key = string(buf + i, KEYSIZE);		//hash 8 bytes of our string
			
			iNum = hash(key);
			//bounds = geneIndex.equal_range(key);		//old indexer
			
			for(giIter = geneIndex[iNum]->begin(), bound = geneIndex[iNum]->end(); giIter!=bound; ++giIter) { //iterate through indices
				mapKey = indexScore.find(*giIter - i);				//find out if index (minus i) is already in map
					
				if(mapKey == indexScore.end()) {		 //if they aren't in map, insert them with value = 1
					indexScore.insert(make_pair(*giIter - i, 1));
				}
				else {				//if they are in map, add 1 to their value
					//put off on this short-circuiting for now
					//if(mapKey->second == 9) {			//if we have 9 hits, declare this index the winner
					//	outf << mapKey->first << endl;
					//	goto beginMap;
					//}
					mapKey->second++;
				}
			}
		}

		//iterate through map to find key with highest value
		for(mapKey = indexScore.begin(); mapKey != indexScore.end(); mapKey++) {		
			if(mapKey->second > LTscore[readNum].quality) {
				LTscore[readNum].index = mapKey->first;
				LTscore[readNum].quality = mapKey->second;
			}
		}			
		readNum++;
	}
			
	//close file
	inf.close();

	return 0;
}

int indexer::commitResults(const std::string resultsFile) {
	ofstream out;

	out.open(resultsFile.c_str());

	vector<score>::iterator beg, end;

	for(beg = LTscore.begin(), end = LTscore.end(); beg != end; beg++) {
		out << beg->index << " " << beg->quality << endl;
	}

	if(out.bad()) {
		cerr << "Error committing results!" << endl;
		return 1;
	}

	return 0;
}
